@@ -13,18 +13,22 @@
 package Net::STOMP::Client;
 use strict;
 use warnings;
-our $VERSION = sprintf("%d.%02d", q$Revision: 1.57 $ =~ /(\d+)\.(\d+)/);
+our $VERSION  = "1.0";
+our $REVISION = sprintf("%d.%02d", q$Revision: 1.82 $ =~ /(\d+)\.(\d+)/);
 
 #
 # used modules
 #
 
 use IO::Socket::INET;
-use List::Util qw(shuffle);
+use List::Util qw(max shuffle);
+use Time::HiRes qw();
 use Net::STOMP::Client::Debug;
 use Net::STOMP::Client::Error;
 use Net::STOMP::Client::Frame;
 use Net::STOMP::Client::IO;
+use Net::STOMP::Client::Peer;
+use Net::STOMP::Client::Protocol qw(FLAG_DIRECTION_C2S FLAG_DIRECTION_S2C);
 
 #
 # Object Oriented definition
@@ -32,8 +36,11 @@ use Net::STOMP::Client::IO;
 
 use Net::STOMP::Client::OO;
 our(@ISA) = qw(Net::STOMP::Client::OO);
-Net::STOMP::Client::OO::methods(qw(uri host port timeout sockopts callbacks session
-				   _io _id _serial _receipts));
+Net::STOMP::Client::OO::methods(qw(
+    uri host port timeout sockopts callbacks peer session
+    client_heart_beat server_heart_beat
+    _io _id _serial _receipts _version
+));
 
 #
 # global variables
@@ -49,18 +56,51 @@ our(
 #                                                                              #
 #---############################################################################
 
-# FIXME: check and use $self->timeout() where this makes sense...
+#
+# return which timeout to use for the given operation
+#
+
+sub _timeout : method {
+    my($self, $operation) = @_;
+    my($timeout);
+
+    $timeout = $self->timeout();
+    if ($timeout) {
+	if (ref($timeout) eq "HASH") {
+	    # hash timeout specified -> use it as is
+	} else {
+	    # scalar timeout specified -> backward compatibility
+	    $timeout = {
+		connect   => $timeout,
+		connected => $timeout,
+		send      => undef,
+	    };
+	}
+    } else {
+	# no timeout specified -> use hard-coded defaults
+	$timeout = {
+	    connect   => undef,
+	    connected => 10,
+	    send      => undef,
+	};
+    }
+    unless ($operation =~ /^(connect|connected|send)$/) {
+	Net::STOMP::Client::Error::repor("%s: unexpected timeout operation: %s",
+					 "Net::STOMP::Client", $operation);
+	return();
+    }
+    return($timeout->{$operation});
+}
 
 #
 # try to connect to the server (socket level only)
 #
 
-sub _new_socket ($$$%) {
-    my($proto, $host, $port, %sockopts) = @_;
-    my($me, $socket);
+sub _new_socket ($$$$%) {
+    my($me, $proto, $host, $port, %sockopts) = @_;
+    my($socket);
 
     # initial checks
-    $me = "Net::STOMP::Client->new()";
     $sockopts{Proto} = "tcp"; # yes, even SSL is TCP...
     $sockopts{PeerAddr} = $host;
     $sockopts{PeerPort} = $port;
@@ -102,16 +142,15 @@ sub _new_socket ($$$%) {
 }
 
 #
-# check and handle the uri opion
+# check and handle the uri option
 #
 # see: http://activemq.apache.org/failover-transport-reference.html
 #
 
-sub _handle_uri ($) {
-    my($string) = @_;
-    my($me, @uris, $uri, $path, $fh, $line, @list, @servers);
+sub _handle_uri ($$) {
+    my($me, $string) = @_;
+    my(@uris, $uri, $path, $fh, $line, @list, @peers);
 
-    $me = "Net::STOMP::Client->new()";
     @uris = ($string);
     while (@uris) {
 	$uri = shift(@uris);
@@ -144,21 +183,55 @@ sub _handle_uri ($) {
 	    unshift(@uris, shuffle(@list));
 	} elsif ($uri =~ /^(tcp|ssl|stomp|stomp\+ssl):\/\/([\w\.\-]+):(\d+)\/?$/) {
 	    # well-formed uri
-	    push(@servers, {
+	    push(@peers, Net::STOMP::Client::Peer->new(
 		proto => $1,
 		host  => $2,
 		port  => $3,
-	    });
+	    ));
 	} else {
 	    Net::STOMP::Client::Error::report("%s: unexpected server uri: %s", $me, $string);
 	    return();
 	}
     }
-    unless (@servers) {
+    unless (@peers) {
 	Net::STOMP::Client::Error::report("%s: empty server uri: %s", $me, $string);
 	return();
     }
-    return(@servers);
+    return(@peers);
+}
+
+#
+# check the version option
+#
+
+sub _check_version ($$) {
+    my($me, $version) = @_;
+    my(@list);
+
+    if (defined($version)) {
+	if (ref($version) eq "ARRAY") {
+	    @list = sort(@$version);
+	} else {
+	    @list = ($version);
+	}
+    } else {
+	# default versions to use
+	@list = qw(1.0 1.1);
+    }
+    unless (@list) {
+	Net::STOMP::Client::Error::report("%s: empty list of STOMP versions given", $me);
+	return();
+    }
+    foreach $version (@list) {
+	unless ($version =~ /^1\.[01]$/) {
+	    Net::STOMP::Client::Error::report("%s: unsupported STOMP version: %s",
+					      $me, $version);
+	    return();
+	}
+    }
+    # modify the given version argument
+    $_[1] = \@list;
+    return(1);
 }
 
 #
@@ -167,14 +240,19 @@ sub _handle_uri ($) {
 
 sub new : method {
     my($class, %data) = @_;
-    my($me, $self, %sockopts, @uris, $uri, @servers, $server, $socket, $io);
+    my($me, $self, %sockopts, $timeout, @uris, $uri, @peers, $peer, $socket, $io, $version);
 
     $me = "Net::STOMP::Client->new()";
+    $version = delete($data{version});
     $self = $class->SUPER::new(%data);
+    _check_version($me, $version) or return();
+    $self->_version($version);
     %sockopts = %{ $self->sockopts() } if $self->sockopts();
     $sockopts{SSL_use_cert} = 1 if $sockopts{SSL_cert_file} or $sockopts{SSL_key_file};
-    $sockopts{Timeout} = $self->timeout()
-	if $self->timeout() and not exists($sockopts{Timeout});
+    unless (exists($sockopts{Timeout})) {
+	$timeout = $self->_timeout("connect");
+	$sockopts{Timeout} = $timeout if $timeout;
+    }
     if ($self->uri()) {
 	if ($self->host()) {
 	    Net::STOMP::Client::Error::report("%s: unexpected server host: %s",
@@ -186,10 +264,13 @@ sub new : method {
 					      $me, $self->port());
 	    return();
 	}
-	@servers = _handle_uri($self->uri());
-	return() unless @servers;
-	# use by default a shorter (but hard-coded) timeout in case we use failover...
-	$sockopts{Timeout} = 1 if @servers > 1 and not exists($sockopts{Timeout});
+	@peers = _handle_uri($me, $self->uri());
+	return() unless @peers;
+	# use by default a shorter timeout in case we use failover...
+	if (@peers > 1 and not $sockopts{Timeout}) {
+	    # it makes no sense to block with failover!
+	    $sockopts{Timeout} = 3;
+	}
     } else {
 	unless ($self->host()) {
 	    Net::STOMP::Client::Error::report("%s: missing server host", $me);
@@ -199,24 +280,32 @@ sub new : method {
 	    Net::STOMP::Client::Error::report("%s: missing server port", $me);
 	    return();
 	}
-	@servers = ({
+	@peers = (Net::STOMP::Client::Peer->new(
 	    proto => grep(/^SSL_/, keys(%sockopts)) ? "ssl" : "tcp",
 	    host  => $self->host(),
 	    port  => $self->port(),
-	});
+	));
     }
-    foreach $server (@servers) {
-	$socket = _new_socket($server->{proto}, $server->{host}, $server->{port}, %sockopts);
-	last if $socket;
+    foreach $peer (@peers) {
+	$socket = _new_socket($me, $peer->proto(), $peer->host(), $peer->port(), %sockopts);
+	if ($socket) {
+	    # keep track of the peer we are connected to
+	    $peer->addr($socket->peerhost()); # real peer address
+	    $self->peer($peer);
+	    last;
+	}
     }
     unless ($socket) {
 	Net::STOMP::Client::Error::report("%s", $Net::STOMP::Client::Error::Message);
 	return();
     }
-    # regardless of what we've been given, always record peer information this way
-    $self->host($socket->peerhost());
-    $self->port($socket->peerport());
     # final setup steps
+    if ($self->uri()) {
+	# keep track of the peer this way too...
+	$peer = $self->peer();
+	$self->host($peer->host());
+	$self->port($peer->port());
+    }
     $io = Net::STOMP::Client::IO->new($socket) or return();
     $self->_io($io);
     unless ($self =~ /\(0x(\w+)\)/) {
@@ -241,14 +330,27 @@ sub send_frame : method {
     my($self, $frame, $timeout) = @_;
     my($done, $receipt);
 
-    # always check the sent frame
-    $frame->check() or return();
-    # keep track of receipts
-    $receipt = $frame->header("receipt");
-    $self->_receipts()->{$receipt}++ if $receipt;
-    # encode the frame and send it
-    $frame->debug(" sending") if $Net::STOMP::Client::Debug::Flags;
-    $done = $self->_io()->send_data($frame->encode(), $timeout);
+    # check that the object is usable
+    Net::STOMP::Client::Error::report("lost connection")
+	unless $self->_io();
+    # handle global send timeout
+    $timeout = $self->_timeout("send") unless defined($timeout);
+    if (ref($frame)) {
+	# always check the sent frame
+	$frame->check(
+	    version   => $self->version(),
+	    direction => FLAG_DIRECTION_C2S,
+	) or return();
+	# keep track of receipts
+	$receipt = $frame->header("receipt");
+	$self->_receipts()->{$receipt}++ if $receipt;
+	# encode the frame and send it
+	$frame->debug(" sending") if $Net::STOMP::Client::Debug::Flags;
+	$done = $self->_io()->send_data($frame->_encode($self->version()), $timeout);
+    } else {
+	# handle the special NOOP frame
+	$done = $self->_io()->send_data($frame, $timeout);
+    }
     return($done) unless $done;
     return($frame);
 }
@@ -259,26 +361,37 @@ sub send_frame : method {
 
 sub receive_frame : method {
     my($self, $timeout) = @_;
-    my($buffer, $done, $frame);
+    my($version, $buffer, $done, $frame);
 
+    # check that the object is usable
+    Net::STOMP::Client::Error::report("lost connection")
+	unless $self->_io();
     # first try to use the current buffer
-    $buffer = $self->_io()->_buffer();
-    $frame = Net::STOMP::Client::Frame::decode($buffer);
+    $version = $self->version();
+    $buffer = $self->_io()->_incoming_buffer();
+    $frame = Net::STOMP::Client::Frame::_decode($buffer, $version);
     return() unless defined($frame);
     unless ($frame) {
 	# not enough data in buffer for a complete frame
 	$done = $self->_io()->receive_data($timeout);
 	return($done) unless $done;
-	$buffer = $self->_io()->_buffer();
+	$buffer = $self->_io()->_incoming_buffer();
 	# try once more with new buffer
-	$frame = Net::STOMP::Client::Frame::decode($buffer);
+	$frame = Net::STOMP::Client::Frame::_decode($buffer, $version);
 	return() unless defined($frame);
-	return(0) unless $frame;
+	unless ($frame) {
+	    # update the buffer (which may have changed)
+	    $self->_io()->_incoming_buffer($buffer);
+	    return(0);
+	}
     }
     # update the buffer (which must have changed)
-    $self->_io()->_buffer($buffer);
+    $self->_io()->_incoming_buffer($buffer);
     # always check the received frame
-    $frame->check() or return();
+    $frame->check(
+	version   => $self->version(),
+	direction => FLAG_DIRECTION_S2C,
+    ) or return();
     # so far so good
     $frame->debug(" received") if $Net::STOMP::Client::Debug::Flags;
     return($frame);
@@ -290,13 +403,16 @@ sub receive_frame : method {
 
 sub wait_for_frames : method {
     my($self, %option) = @_;
-    my($timeout, $callback, $limit, $frame, $result);
+    my($callback, $timeout, $limit, $timeleft, $frame, $result, $now);
 
     $callback = $option{callback};
     $timeout = $option{timeout};
-    $limit = time() + $timeout if defined($timeout);
+    if (defined($timeout)) {
+	$limit = Time::HiRes::time() + $timeout;
+	$timeleft = $timeout;
+    }
     while (1) {
-	$frame = $self->receive_frame($timeout);
+	$frame = $self->receive_frame($timeleft);
 	return() unless defined($frame);
 	if ($frame) {
 	    # we always call first the per-command callback
@@ -311,7 +427,12 @@ sub wait_for_frames : method {
 		return($frame);
 	    }
 	}
-	return(0) if defined($timeout) and time() > $limit;
+	# we check if we exceeded the timeout
+	if (defined($timeout)) {
+	    $now = Time::HiRes::time();
+	    return(0) if $now >= $limit;
+	    $timeleft = $limit - $now;
+	}
     }
     # not reached...
     die("ooops!");
@@ -330,6 +451,28 @@ sub wait_for_receipts : method {
     return($self->wait_for_frames(%option));
 }
 
+#
+# return the socket used
+#
+
+sub socket : method {
+    my($self) = @_;
+
+    return($self->_io()->_socket());
+}
+
+#
+# return the negotiated protocol version (or undef if not negotiated yet)
+#
+
+sub version : method {
+    my($self) = @_;
+    my($version);
+
+    $version = $self->_version();
+    return($version) unless ref($version);
+    return(undef);
+}
 #
 # return a universal pseudo-unique id to be used in receipts and transactions
 #
@@ -351,6 +494,26 @@ sub receipts : method {
     return(keys(%{ $self->_receipts() }));
 }
 
+#
+# return the last time we received data
+#
+
+sub last_received : method {
+    my($self) = @_;
+
+    return($self->_io() ? $self->_io()->_last_read() : undef);
+}
+
+#
+# return the last time we sent data
+#
+
+sub last_sent : method {
+    my($self) = @_;
+
+    return($self->_io() ? $self->_io()->_last_write() : undef);
+}
+
 #
 # object destructor
 #
@@ -421,12 +584,60 @@ sub _unexpected_frame ($$) {
 }
 
 #
-# default CONNECTED frame callback: keep track of the session identifier
+# default CONNECTED frame callback: keep track of the negotiated
+# protocol version as well as the session identifier
 #
 
 sub _default_connected_callback ($$) {
     my($self, $frame) = @_;
+    my($value);
 
+    # STOMP 1.1: handle version negotiation
+    $value = $frame->header("version");
+    if ($value) {
+	if (grep($_ eq $value, @{ $self->_version() })) {
+	    $self->_version($value);
+	} else {
+	    _unexpected_frame($frame, "unexpected STOMP version: $value");
+	    return();
+	}
+    } else {
+	$value = "1.0";
+	if (grep($_ eq $value, @{ $self->_version() })) {
+	    $self->_version($value);
+	} else {
+	    _unexpected_frame($frame, "server only supports STOMP $value");
+	    return();
+	}
+    }
+    # STOMP 1.1: handle heart-beat negotiation
+    $value = $frame->header("heart-beat");
+    if ($value) {
+	if ($value =~ /^(\d+),(\d+)$/) {
+	    if ($self->client_heart_beat() and $2) {
+		# negotiated
+		$self->client_heart_beat(max($self->client_heart_beat(), $2 / 1000.0));
+	    } else {
+		# disabled
+		$self->client_heart_beat(0);
+	    }
+	    if ($self->server_heart_beat() and $1) {
+		# negotiated
+		$self->server_heart_beat(max($self->server_heart_beat(), $1 / 1000.0));
+	    } else {
+		# disabled
+		$self->server_heart_beat(0);
+	    }
+	} else {
+	    _unexpected_frame($frame, "unexpected heart-beat setting: $value");
+	    return();
+	}
+    } else {
+	# missing value -> disable
+	$self->client_heart_beat(0);
+	$self->server_heart_beat(0);
+    }
+    # keep track of session id too
     unless ($self->session()) {
 	$self->session($frame->header("session"));
 	return($self);
@@ -443,7 +654,6 @@ $Callback{CONNECTED} = \&_default_connected_callback;
 sub _default_error_callback ($$) {
     my($self, $frame) = @_;
 
-    # FIXME: what shall we do with the message body?
     _unexpected_frame($frame, $frame->header("message"));
     return();
 }
@@ -540,33 +750,47 @@ sub _check_invocation : method {
 
 sub connect : method {
     my($self, %option) = @_;
-    my($me, $frame, $session);
+    my($me, $frame, $timeout, @list, $session);
 
     $me = "Net::STOMP::Client->connect()";
     # debug
-    Net::STOMP::Client::Debug::report(Net::STOMP::Client::Debug::API, "%s()", $me);
+    Net::STOMP::Client::Debug::report(Net::STOMP::Client::Debug::API, "%s", $me);
     # must be called with a hash
     unless (scalar(@_) % 2) {
-	Net::STOMP::Client::Error::report("%s(): wrong invocation", $me);
+	Net::STOMP::Client::Error::report("%s: wrong invocation", $me);
 	return();
     }
     # must be called before connection
     if ($self->session()) {
-	Net::STOMP::Client::Error::report("%s(): already connected with session: %s",
+	Net::STOMP::Client::Error::report("%s: already connected with session: %s",
 					  $me, $self->session());
 	return();
     }
     # so far so good
-    $option{login} = "" unless defined($option{login});
-    $option{passcode} = "" unless defined($option{passcode});
+    $timeout = delete($option{timeout});
+    if (grep($_ eq "1.1", @{ $self->_version() })) {
+	# supply additional STOMP 1.1 headers
+	$option{"accept-version"} = join(",", @{ $self->_version() })
+	    unless defined($option{"accept-version"});
+	$option{host} = $self->host()
+	    unless defined($option{host});
+	unless (defined($option{"heart-beat"})) {
+	    @list = ();
+	    # the protocol expects values in milliseconds...
+	    push(@list, int(($self->client_heart_beat() || 0) * 1000.0 + 0.5));
+	    push(@list, int(($self->server_heart_beat() || 0) * 1000.0 + 0.5));
+	    $option{"heart-beat"} = join(",", @list)
+		unless "@list" eq "0 0";
+	}
+    }
     $frame = Net::STOMP::Client::Frame->new(
-        command => "CONNECT",
+	command => "CONNECT",
 	headers => \%option,
     );
-    $self->send_frame($frame) or return();
+    $self->send_frame($frame, $timeout) or return();
     $session = $self->wait_for_frames(
-        callback => sub { return($self->session()) },
-	timeout  => 10,
+	callback => sub { return($self->session()) },
+	timeout  => $self->_timeout("connected"),
     );
     return($self) if $session;
     Net::STOMP::Client::Error::report("Net::STOMP::Client->connect(): %s",
@@ -575,21 +799,25 @@ sub connect : method {
 }
 
 #
-# disconnect from server (no options/headers supported)
+# disconnect from server
 #
 
 sub disconnect : method {
-    my($self) = @_;
-    my($frame);
+    my($self, %option) = @_;
+    my($frame, $timeout);
 
     $self->_check_invocation(scalar(@_)) or return();
+    $timeout = delete($option{timeout});
     # send a DISCONNECT frame
-    $frame = Net::STOMP::Client::Frame->new(command => "DISCONNECT");
-    $self->send_frame($frame) or return();
+    $frame = Net::STOMP::Client::Frame->new(
+	command => "DISCONNECT",
+	headers => \%option,
+    );
+    $self->send_frame($frame, $timeout) or return();
     # additional bookkeeping
     $self->session("");
-    # FIXME: close the socket and mark the object as not usable anymore
-    # FIXME: warning if the buffer is not empty, we may have unprocessed frames...
+    $self->peer(undef);
+    $self->_io(undef);
     return($self);
 }
 
@@ -599,15 +827,16 @@ sub disconnect : method {
 
 sub subscribe : method {
     my($self, %option) = @_;
-    my($frame);
+    my($frame, $timeout);
 
     $self->_check_invocation(scalar(@_)) or return();
+    $timeout = delete($option{timeout});
     # send a SUBSCRIBE frame
     $frame = Net::STOMP::Client::Frame->new(
-        command => "SUBSCRIBE",
+	command => "SUBSCRIBE",
 	headers => \%option,
     );
-    $self->send_frame($frame) or return();
+    $self->send_frame($frame, $timeout) or return();
     return($self);
 }
 
@@ -617,15 +846,16 @@ sub subscribe : method {
 
 sub unsubscribe : method {
     my($self, %option) = @_;
-    my($frame);
+    my($frame, $timeout);
 
     $self->_check_invocation(scalar(@_)) or return();
+    $timeout = delete($option{timeout});
     # send an UNSUBSCRIBE frame
     $frame = Net::STOMP::Client::Frame->new(
-        command => "UNSUBSCRIBE",
+	command => "UNSUBSCRIBE",
 	headers => \%option,
     );
-    $self->send_frame($frame) or return();
+    $self->send_frame($frame, $timeout) or return();
     return($self);
 }
 
@@ -635,19 +865,18 @@ sub unsubscribe : method {
 
 sub send : method {
     my($self, %option) = @_;
-    my($frame, $body);
+    my($frame, $timeout, $body);
 
     $self->_check_invocation(scalar(@_)) or return();
-    # the body is given via %option, we move it out
-    $body = $option{body};
-    delete($option{body});
+    $timeout = delete($option{timeout});
+    $body = delete($option{body});
     # send a SEND frame
     $frame = Net::STOMP::Client::Frame->new(
-        command => "SEND",
+	command => "SEND",
 	headers => \%option,
     );
     $frame->body($body) if defined($body);
-    $self->send_frame($frame) or return();
+    $self->send_frame($frame, $timeout) or return();
     return($self);
 }
 
@@ -657,20 +886,58 @@ sub send : method {
 
 sub ack : method {
     my($self, %option) = @_;
-    my($frame);
+    my($frame, $timeout, $value);
 
     $self->_check_invocation(scalar(@_)) or return();
-    # the message id can be given via a frame object in %option
+    $timeout = delete($option{timeout});
+    # we can optionally give a MESSAGE frame here
     if ($option{frame}) {
-	$option{"message-id"} = $option{frame}->header("message-id");
+	# we are careful in case the frame is not a MESSAGE frame...
+	$value = $option{frame}->header("message-id");
+	$option{"message-id"} = $value if defined($value);
+	$value = $option{frame}->header("subscription");
+	$option{"subscription"} = $value if defined($value);
 	delete($option{frame});
     }
     # send an ACK frame
     $frame = Net::STOMP::Client::Frame->new(
-        command => "ACK",
+	command => "ACK",
 	headers => \%option,
     );
-    $self->send_frame($frame) or return();
+    $self->send_frame($frame, $timeout) or return();
+    return($self);
+}
+
+#
+# acknowledge the rejection of a message
+#
+
+sub nack : method {
+    my($self, %option) = @_;
+    my($frame, $timeout, $value);
+
+    $self->_check_invocation(scalar(@_)) or return();
+    if ($self->version() eq "1.0") {
+	Net::STOMP::Client::Error::report("Net::STOMP::Client->nack(): %s %s",
+					  "not supported for STOMP", $self->version());
+	return();
+    }
+    $timeout = delete($option{timeout});
+    # we can optionally give a MESSAGE frame here
+    if ($option{frame}) {
+	# we are careful in case the frame is not a MESSAGE frame...
+	$value = $option{frame}->header("message-id");
+	$option{"message-id"} = $value if defined($value);
+	$value = $option{frame}->header("subscription");
+	$option{"subscription"} = $value if defined($value);
+	delete($option{frame});
+    }
+    # send an NACK frame
+    $frame = Net::STOMP::Client::Frame->new(
+	command => "NACK",
+	headers => \%option,
+    );
+    $self->send_frame($frame, $timeout) or return();
     return($self);
 }
 
@@ -680,15 +947,16 @@ sub ack : method {
 
 sub begin : method {
     my($self, %option) = @_;
-    my($frame);
+    my($frame, $timeout);
 
     $self->_check_invocation(scalar(@_)) or return();
+    $timeout = delete($option{timeout});
     # send a BEGIN frame
     $frame = Net::STOMP::Client::Frame->new(
-        command => "BEGIN",
+	command => "BEGIN",
 	headers => \%option,
     );
-    $self->send_frame($frame) or return();
+    $self->send_frame($frame, $timeout) or return();
     return($self);
 }
 
@@ -698,15 +966,16 @@ sub begin : method {
 
 sub commit : method {
     my($self, %option) = @_;
-    my($frame);
+    my($frame, $timeout);
 
     $self->_check_invocation(scalar(@_)) or return();
+    $timeout = delete($option{timeout});
     # send a COMMIT frame
     $frame = Net::STOMP::Client::Frame->new(
-        command => "COMMIT",
+	command => "COMMIT",
 	headers => \%option,
     );
-    $self->send_frame($frame) or return();
+    $self->send_frame($frame, $timeout) or return();
     return($self);
 }
 
@@ -716,18 +985,54 @@ sub commit : method {
 
 sub abort : method {
     my($self, %option) = @_;
-    my($frame);
+    my($frame, $timeout);
 
     $self->_check_invocation(scalar(@_)) or return();
+    $timeout = delete($option{timeout});
     # send an ABORT frame
     $frame = Net::STOMP::Client::Frame->new(
-        command => "ABORT",
+	command => "ABORT",
 	headers => \%option,
     );
-    $self->send_frame($frame) or return();
+    $self->send_frame($frame, $timeout) or return();
     return($self);
 }
 
+#
+# send an empty/noop frame (= a single newline byte)
+#
+
+sub noop : method {
+    my($self, %option) = @_;
+    my($frame, $timeout);
+
+    $self->_check_invocation(scalar(@_)) or return();
+    $timeout = delete($option{timeout});
+    # there is no NOOP frame (yet) so we simply send a newline
+    $frame = "\n";
+    $self->send_frame($frame, $timeout) or return();
+    return($self);
+}
+
+#
+# idem but only if needed wrt heart-beat settings
+#
+
+sub beat : method {
+    my($self, %option) = @_;
+    my($delta, $sent);
+
+    $self->_check_invocation(scalar(@_)) or return();
+    # check if client heart-beats are expected
+    $delta = $self->client_heart_beat();
+    return($self) unless $delta;
+    # check the last time we sent data
+    $sent = $self->last_sent();
+    return($self) if Time::HiRes::time() - $sent < $delta / 2;
+    # send a noop frame
+    return($self->noop(%option));
+}
+
 1;
 
 __END__
@@ -796,6 +1101,11 @@ supported:
 
 =over
 
+=item C<version>
+
+the STOMP version to use (string) or versions to use (reference to a
+list of strings); this defaults to the list of all supported versions
+
 =item C<uri>
 
 the Uniform Resource Identifier (URI) specifying where the STOMP
@@ -813,7 +1123,8 @@ the port number of the STOMP service
 
 =item C<timeout>
 
-a timeout value to be used in some I/O operation (experimental)
+the maximum time (in seconds) for various operations, see the L</"TIMEOUTS">
+section for more information
 
 =item C<sockopts>
 
@@ -824,9 +1135,15 @@ IO::Socket::INET->new() or IO::Socket::SSL->new()
 
 a hash of code references that will be called on each received frame
 
-=item C<session>
+=item C<client_heart_beat>
 
-the session identifier, once connected
+the desired client-side heart-beat setting, see the L</"HEART-BEATING">
+section for more information
+
+=item C<server_heart_beat>
+
+the desired server-side heart-beat setting, see the L</"HEART-BEATING">
+section for more information
 
 =back
 
@@ -868,6 +1185,42 @@ private key
 
 For more information, see L<IO::Socket::SSL>.
 
+=head1 TIMEOUTS
+
+By default, when sending STOMP frames, the module waits until the
+frame indeed has been sent (from the socket point of view). In case
+the server is stuck or unusable, the module can therefore hang.
+
+When creating the Net::STOMP::Client object, you can pass a C<timeout>
+attribute to better control how certain operations handle timeouts.
+
+This attribute should contain a reference to hash with the following
+keys:
+
+=over
+
+=item connect
+
+TCP-level timeout that will be given to the underlying
+L<IO::Socket::INET> or L<IO::Socket::SSL> object (default: none)
+
+=item connected
+
+timeout used while waiting for the initial CONNECTED frame from the
+broker (default: 10)
+
+=item send
+
+timeout used while trying to send any frame (default: none)
+
+=back
+
+All values are in seconds. No timeout means wait until the operation
+succeeds.
+
+As a shortcut, the C<timeout> attribute can also be a scalar. In this
+case, only the C<connect> and C<connected> operations use this value.
+
 =head1 STOMP METHODS
 
 With a Net::STOMP::Client object, the following methods can be used to
@@ -900,6 +1253,11 @@ send a message somewhere
 
 acknowledge the reception of a message
 
+=item nack()
+
+acknowledge the rejection of a message
+(STOMP 1.1 only)
+
 =item begin()
 
 begin/start a transaction
@@ -932,16 +1290,45 @@ C<body>: holds the body of the message to be sent
 
 =item ack()
 
-C<frame>: holds the frame object to acknowledge (its message-id will be used)
+C<frame>: holds the MESSAGE frame object to ACK
+
+=item nack()
+
+C<frame>: holds the MESSAGE frame object to NACK
 
 =back
 
+Finally, all methods support a C<timeout> option that will be given to
+the send_frame() method called internally to send the crafted frame.
+
 =head1 OTHER METHODS
 
 In addition to the STOMP methods, the following ones are also available:
 
 =over
 
+=item new(OPTIONS)
+
+return a new Net::STOMP::Client object
+
+=item peer()
+
+return a Net::STOMP::Client::Peer object containing information about
+the connected STOMP server
+
+=item socket()
+
+return the file handle of the socket connecting the client and the server
+
+=item session()
+
+return the session identifier if connected or undef otherwise
+
+=item version()
+
+return the STOMP version that has been negotiated between the client
+and the server or undef if the negotiation did not take place yet
+
 =item uuid()
 
 return a universal pseudo-unique identifier to be used for instance in
@@ -949,7 +1336,7 @@ receipts and transactions
 
 =item receipts()
 
-return the list of not-yet-received receipts, see the L</RECEIPTS>
+return the list of not-yet-received receipts, see the L</"RECEIPTS">
 section for more information
 
 =item wait_for_frames()
@@ -961,6 +1348,11 @@ information
 
 wait for all receipts to be received, using wait_for_frames() underneath
 
+=item noop([timeout => TIMEOUT])
+
+send an empty/noop frame i.e. a single newline byte, using
+send_frame() underneath
+
 =back
 
 =head1 CALLBACKS
@@ -984,6 +1376,21 @@ Here is an example with a callback counting the messages received:
       return($self);
   });
 
+Here are the methods that can be used to get or set these per-command
+callbacks:
+
+=over
+
+=item connected_callback([SUBREF])
+
+=item error_callback([SUBREF])
+
+=item message_callback([SUBREF])
+
+=item receipt_callback([SUBREF])
+
+=back
+
 These callbacks are somehow global and it is good practice not to
 change them during a session. If you do not need a global message
 callback, you can supply the dummy:
@@ -1075,6 +1482,55 @@ Here is an example using transactions:
   # commit the transaction
   $stomp->commit(transaction => $tid);
 
+=head1 HEART-BEATING
+
+STOMP 1.1 defines how each end of a STOMP connection can check if the
+other end is alive. To support heart-beating, this module provides the
+following methods:
+
+=over
+
+=item last_received()
+
+return the time at which data was last received, i.e. read from the
+network socket
+
+=item last_sent()
+
+return the time at which data was last sent, i.e. written to the
+network socket
+
+=item client_heart_beat()
+
+(after having received the CONNECTED frame) return the negotiated
+client-side heart-beat setting
+
+=item server_heart_beat()
+
+(after having received the CONNECTED frame) return the negotiated
+server-side heart-beat setting
+
+=item beat([timeout => TIMEOUT])
+
+send a noop frame (using the noop() method) unless the last sent time
+is recent enough with regard to the client heart-beat setting
+
+=back
+
+For consistency with other Perl modules (for instance Time::HiRes),
+time is always expressed as a fractional number of seconds.
+
+To use heart-beating, the client must specify the desired
+C<client_heart_beat> and/or C<server_heart_beat> attributes when
+invoking the new() method. Then, once the CONNECTED frame has been
+received, it can get the negotiated values with the methods above.
+
+To prove that it is alive, the client just needs to call the beat()
+method.
+
+To check if the server is alive, the client just needs to compare
+what is returned by the last_received() and server_heart_beat() methods.
+
 =head1 LOW-LEVEL API
 
 It should be enough to use the high-level API and use, for instance,
@@ -1087,6 +1543,10 @@ You can also use:
 
 =over
 
+=item $stomp->dispatch_frame(FRAME)
+
+dispatch one received frame by calling the appropriate callback
+
 =item $stomp->send_frame(FRAME, TIMEOUT)
 
 try to send the given frame object within the given TIMEOUT, or
@@ -1101,9 +1561,11 @@ TIMEOUT is undef
 
 =head1 COMPATIBILITY
 
-This module implements the version 1.0 of the protocol (see
-L<http://stomp.codehaus.org/Protocol>) as well as well known
-extensions for JMS, AMQP, ActiveMQ and RabbitMQ.
+This module implements the versions C<1.0> (see
+L<http://stomp.github.com/stomp-specification-1.0.html>) and C<1.1>
+(see L<http://stomp.github.com/stomp-specification-1.1.html>) of the
+protocol as well as well known extensions for JMS, AMQP, ActiveMQ and
+RabbitMQ.
 
 It has been successfully tested against both ActiveMQ and RabbitMQ
 brokers.
@@ -1112,9 +1574,13 @@ brokers.
 
 L<IO::Socket::INET>,
 L<IO::Socket::SSL>,
+L<Net::STOMP::Client::Error>,
 L<Net::STOMP::Client::Frame>,
-L<Net::STOMP::Client::Error>.
+L<Net::STOMP::Client::Peer>,
+L<Time::HiRes>.
 
 =head1 AUTHOR
 
 Lionel Cons L<http://cern.ch/lionel.cons>
+
+Copyright CERN 2010-2011
